"""Sheet behavior definitions
Each top level event is a series of actions leading to
the completion of an event.


"""

from .apoptosis_events import apoptosis
from .delamination_events import constriction
from .basic_events import division, type1_transition, face_elimination
