
# THIS FILE IS GENERATED FROM tyssue SETUP.PY
#
short_version = '0.6.7'
full_version = '0.6.7'
git_revision = 'fec8c8cc847f478c83656dffa850224e769b51db'
release = True
if release:
    version = full_version
else:
    version = short_version
